%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: lstm_fit_sgd.m,v 1.2 2017/06/28 08:49:10 mechanoid Exp mechanoid $


% - - - - - - - - - - - - - - - - 
% обучение LSTM методом градиентного спуска (стохастический)
function [W,er_hist]=lstm_fit_sgd(W,X,y,Xc,yc),
   min_err =  0.07 ; % порог допустимой ошибки обучения
   max_iter = 12e3 ; % максимальное число эпох
   er_win_size = 3 ;  % размер окна контроля за ошибкой
   trsp = 0.01 ; % параметр скорости обучения

   batch_par = 0.25 ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = 1 ;% печатаем значение ошибки каждые print_step эпох 

   num_examples=size(y,2) ; % количество объектов в учебном наборе
   batch_size = max(2,round(num_examples*batch_par)) ; % количество учебных примеров в одном пакете 
   
   printf("[i] обучение методом стохастического градиентного спуска...\n");
 
   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   W0 = W ;

   for iter=1:max_iter,

     er_hist(end+1) =  lstm_loss(Xc,W,yc) ; % вычисляем ошибку сети на контрольном наборе

     if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
       printf("      %i/%i : %f / %f   , %f     \r",iter,max_iter,er_hist(end),min_err,trsp ) ; 
     endif
     
     % if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения

     if er_hist(end) < min_err, break; endif
 
     % batch_par=0.01 ;
     de = er_hist(end) - er_hist(end-1) ; % изменение ошибки
     if( de > 0 ), % ошибка выросла  
	p=1.0/(1.0+exp(-de)); % вероятность возврата старого состояния в случае увеличения ошибки
        if (rand() < p), 
	   %printf("\n[i] ошибка выросла, возврат старого состояния с вероятностью %f\n",p);
	    W=W0; 
	    er_hist(end)=[];
	 endif
      endif
      W0 = W ;

     % случайный пакет из учебных примеров
     batch_idx = randperm(num_examples)(1:batch_size); 
      
      for iter2=batch_idx,
	 Xb = X(:,:,iter2) ;
	 yb= y(:,iter2) ;
	 dW = lstm_gradient(Xb,W,yb) ; % вычисляем значения градиента 
	 % dW = lstm_weight_norm(dW) ; % нормируем градиент
	 W = lstm_weight_correct(W,dW,trsp) ; % изменяем веса
 
     endfor

   endfor

   printf("[i]   %i/%i : %f / %f   , %f     \n",iter,max_iter,er_hist(end),min_err,trsp ) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction



% % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% function retval=trn_speed(s,er),
%    k=1.01; % коэффициент допустимого роста ошибки
%    ri=1.01; % коэффициент увеличения скорости
%    rd=0.99; % коэффициент уменьшения скорости
% 
%    d = er(end) > k*er(end-1) ;
% 
%    if( d > 0 ),  % если существенный рост ошибки
%      s*=rd; % то уменьшаем скорость обучения
%    else 
%      s=min(s*ri,1.0); % иначе увеличиваем скорость обучения
%    endif
%    retval=s ;
% endfunction


      % trsp=trn_speed(trsp,er_hist); % скорость обучения (изменения весов)




