%!octave 
%
%
%
%       OS : GNU/Linux 4.2.5
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: lstm_gradient.m,v 1.13 2017/06/28 13:51:13 mechanoid Exp mechanoid $

% - - - - - - - - - - - - - - - - 
function retval = lstm_gradient(X,W,y),
   % процедура вычисления градиента BPTT - Backpropagation through time
   [ sgf,sgi,sgo,sZ,sC,sY ] = lstm_bptt_forward(X,W) ; % прямой проход
   er = lstm_bptt_backward(W,y,sgf,sgi,sgo,sZ,sC,sY) ; % обратный проход
   dW = lstm_bptt_gradient(X,W,sC,sY,er) ; % вычисляем значения градиента для всех весов
   retval = dW ;
endfunction

% - - - - - - - - - - - - - - - - 
function [ sgf,sgi,sgo,sZ,sC,sY ] = lstm_bptt_forward(X,W),
   n = size(X,1) ; % размер одного элемента последовательности
   m = size(X,2) ; % длинна последовательности
   k = size(X,3) ; % количество примеров

   ys = size( W{1}.mem.b, 1 ) ; % размер слоя lstm

   Y = zeros(ys,k) ; % начальное значение выходов 
   C = zeros(ys,k) ; % начальное состояние памяти

   sgf = {} ; % история состояний гейта памяти 
   sgi = {} ; % история состояний гейта обновления
   sgo = {} ; % история состояний гейта выхода
   sZ  = {} ; % история состояний обновлений памяти
   sC  = {} ; % история состояний памяти
   sY  = {} ; % история выходов скрытого слоя

   for t = 1:m,
      Xt = reshape( X(:,t,:) , [n,k] ) ; % элементы последовательностей номер t для всех примеров 
 
      sZt = W{1}.mem.x * Xt + W{1}.mem.y * Y + W{1}.mem.b ;
      Z = tanh( sZt ) ;
      sZ{t} = sZt ;

      sgft = W{1}.forget.x * Xt + W{1}.forget.y * Y + C .* W{1}.forget.c + W{1}.forget.b ; % гейт памяти (forget)
      gf = sigmoid(  sgft ) ; % гейт памяти (forget)
      sgf{t} = sgft ;

      sgit = W{1}.input.x  * Xt + W{1}.input.y  * Y + C .* W{1}.input.c + W{1}.input.b ; % гейт обновления (input)
      gi = sigmoid( sgit  ) ; % гейт обновления (input)
      sgi{t} = sgit ;
  
      C = Z .* gi + C .* gf ; % обновляем память
      sC{t} = C ;
	
      sgot = W{1}.output.x  * Xt + W{1}.output.y  * Y + C .* W{1}.output.c + W{1}.output.b ; % гейт выхода (output) 
      go = sigmoid( sgot ) ; % гейт выхода (output) 
      sgo{t} = sgot ;

      Y = tanh( C ) .* go ; % выход 
      sY{t} = Y ;
   endfor
  
endfunction


% - - - - - - - - - - - - - - - - 
function retval = lstm_bptt_backward(W,y,sgf,sgi,sgo,sZ,sC,sY),
   % обратный проход
   % для каждого примера:
   % вычисляем ошибку выходного слоя
   % вычисляем ошибку скрытого слоя для последнего элемента входной последовательности
   % последовательно, в обратном порядке, вычисляем ошибки скрытого слоя для всех элементов входной последовательности
   
   % n = size(X,1) ; % размер одного элемента последовательности
   % m = size(X,2) ; % длинна последовательности
   % k = size(X,3) ; % количество примеров
   m = length(sC) ; % длинна последовательности


   t = m ;

   er{1}{t}.y = sY{t} - y ;
   er{1}{t}.o = er{1}{t}.y .* tanh( sC{t} ) .* sigmoid_drv( sgo{t} ) ;
   er{1}{t}.c = er{1}{t}.y .* sigmoid(sgo{t}) .* tanh_drv( sC{t} ) ...
		+ er{1}{t}.o .* W{1}.output.c ;
 
   er{1}{t}.f = er{1}{t}.c .* sC{t-1} .* sigmoid_drv( sgf{t} ) ;
   er{1}{t}.i = er{1}{t}.c .* tanh( sZ{t} ) .* sigmoid_drv( sgi{t} ) ;
   er{1}{t}.z = er{1}{t}.c .* sigmoid(sgi{t}) .* tanh_drv( sZ{t} ) ;

   for t = m-1:-1:2, 

     er{1}{t}.y = sY{t} - y ...
		  + transpose( W{1}.mem.y )    * er{1}{t+1}.z ...
		  + transpose( W{1}.forget.y ) * er{1}{t+1}.f ...
		  + transpose( W{1}.output.y ) * er{1}{t+1}.o ...
		  + transpose( W{1}.input.y )  * er{1}{t+1}.i ; 

      er{1}{t}.o = er{1}{t}.y .* tanh( sC{t} ) .* sigmoid_drv( sgo{t} ) ;

      er{1}{t}.c = er{1}{t}.y .* sigmoid(sgo{t}) .* tanh_drv( sC{t} ) ...
		  + er{1}{t}.o   .* W{1}.output.c ...
		  + er{1}{t+1}.i .* W{1}.input.c  ...
		  + er{1}{t+1}.f .* W{1}.forget.c ...
		  + er{1}{t+1}.c .* sigmoid( sgf{t+1} )  ;
 
      er{1}{t}.f = er{1}{t}.c .* sC{t-1} .* sigmoid_drv( sgf{t} ) ;
 
      er{1}{t}.i = er{1}{t}.c .* tanh( sZ{t} ) .* sigmoid_drv( sgi{t} ) ;
 
      er{1}{t}.z = er{1}{t}.c .* sigmoid(sgi{t}) .* tanh_drv( sZ{t} ) ;

   endfor

   t = 1 ;

   er{1}{t}.y = sY{t} - y  ...
	       + transpose( W{1}.mem.y )    * er{1}{t+1}.z ...
	       + transpose( W{1}.forget.y ) * er{1}{t+1}.f ...
	       + transpose( W{1}.output.y ) * er{1}{t+1}.o ...
	       + transpose( W{1}.input.y )  * er{1}{t+1}.i ; 

   er{1}{t}.o = er{1}{t}.y .* tanh( sC{t} ) .* sigmoid_drv( sgo{t} ) ;

   er{1}{t}.c = er{1}{t}.y .* sigmoid(sgo{t}) .* tanh_drv( sC{t} ) ...
	       + er{1}{t}.o   .* W{1}.output.c ...
	       + er{1}{t+1}.i .* W{1}.input.c  ...
	       + er{1}{t+1}.f .* W{1}.forget.c ...
	       + er{1}{t+1}.c .* sigmoid( sgf{t+1} )  ;

   er{1}{t}.f = zeros( size( sC{t} ) ) ;

   er{1}{t}.i = er{1}{t}.c .* tanh( sZ{t} ) .* sigmoid_drv( sgi{t} ) ;

   er{1}{t}.z = er{1}{t}.c .* sigmoid(sgi{t}) .* tanh_drv( sZ{t} ) ;

%    er{1}{t}.x = transpose(W{1}.mem.x)*er{1}{t}.z ...
%           + transpose(W{1}.input.x)*er{1}{t}.i ...
%           + transpose(W{1}.forget.x)*er{1}{t}.f ...
%           + transpose(W{1}.output.o)*er{1}{t}.o ;

   retval = er ;

endfunction


% - - - - - - - - - - - - - - - - 
function retval = lstm_bptt_gradient(X,W,sC,sY,er),  % вычисляем значения градиента для всех весов
   % вычисляем значения градиента для всех весов

   n = size(X,1) ; % размер одного элемента последовательности
   m = size(X,2) ; % длинна последовательности
   k = size(X,3) ; % количество примеров

   dW = {} ;

   dW{1}.mem.x = 0.0 ;
   dW{1}.input.x = 0.0 ;
   dW{1}.forget.x = 0.0 ;
   dW{1}.output.x = 0.0 ;

   dW{1}.mem.y = 0.0 ;
   dW{1}.input.y = 0.0 ;
   dW{1}.forget.y = 0.0 ;
   dW{1}.output.y = 0.0 ;

   dW{1}.input.c = 0.0 ;
   dW{1}.forget.c = 0.0 ;
   dW{1}.output.c = 0.0 ;

   dW{1}.mem.b = 0.0 ;
   dW{1}.input.b = 0.0 ;
   dW{1}.forget.b = 0.0 ;
   dW{1}.output.b = 0.0 ;

   for t = 1:m,
      Xt = reshape( X(:,t,:) , [n,k] ) ; % элементы последовательностей номер t для всех примеров 
     
      dW{1}.mem.x += er{1}{t}.z * transpose( Xt ) ;
      dW{1}.input.x += er{1}{t}.i * transpose( Xt ) ;
      dW{1}.forget.x += er{1}{t}.f * transpose( Xt ) ;
      dW{1}.output.x += er{1}{t}.o * transpose( Xt ) ;
   endfor

   for t = 1:m-1,
      dW{1}.mem.y += er{1}{t+1}.z * transpose( sY{t} ) ;
      dW{1}.input.y += er{1}{t+1}.i * transpose( sY{t} ) ;
      dW{1}.forget.y += er{1}{t+1}.f * transpose( sY{t} ) ;
      dW{1}.output.y += er{1}{t+1}.o * transpose( sY{t} ) ;
   endfor

   for t = 1:m-1,
      dW{1}.input.c  += er{1}{t+1}.i .* sC{t}  ;
      dW{1}.forget.c += er{1}{t+1}.f .* sC{t}  ;
   endfor

   for t = 1:m,
      dW{1}.output.c += er{1}{t}.o .* sC{t}  ;
   endfor

   for t = 1:m,
      dW{1}.mem.b += er{1}{t}.z ;
      dW{1}.input.b += er{1}{t}.i ;
      dW{1}.forget.b += er{1}{t}.f ;
      dW{1}.output.b += er{1}{t}.o ;
   endfor

   dW{1}.input.c  = sum( dW{1}.input.c, 2 ) ;
   dW{1}.forget.c = sum( dW{1}.forget.c, 2 ) ;
   dW{1}.output.c = sum( dW{1}.output.c, 2 ) ;

   dW{1}.mem.b    = sum( dW{1}.mem.b , 2 ) ;
   dW{1}.input.b  = sum( dW{1}.input.b , 2 ) ;
   dW{1}.forget.b = sum( dW{1}.forget.b , 2 ) ;
   dW{1}.output.b = sum( dW{1}.output.b , 2 ) ;


% dW_z = sum_t  er_z * x
% dW_i = sum_t  er_i * x
% dW_f = sum_t  er_f * x
% dW_o = sum_t  er_o * x
% 
% dR_z = sum_t  er_z * y
% dR_i = sum_t  er_i * y
% dR_f = sum_t  er_f * y
% dR_o = sum_t  er_o * y
% 
% dp_i = sum_t c .* er_i
% dp_f = sum_t c .* er_f
% dp_o = sum_t c .* er_o

% db_z = sum_t  er_z 
% db_i = sum_t  er_i 
% db_f = sum_t  er_f 
% db_o = sum_t  er_o 


   retval = dW ;
endfunction
 

% - - - - - - - - - - - - - - - - 
function retval = sigmoid_drv(x),
   retval = exp(-x)./((1.0+exp(-x)).^2);
   % b = 2.0 ; a = 10.0 ; retval=b*a*exp(-x.*a)./((1+exp(-x.*a)).^2);
endfunction

% % - - - - - - - - - - - - - - - - 
% function retval = sigmoid( x ),
%    retval = 1.0 ./ ( 1.0 + exp(-x) ) ;
%    % b=2.0; a=10.0; retval=b./(1.0+exp(-x.*a));
% endfunction

function retval = tanh_drv(x),
   retval = 1.0 ./ (cosh(x).^2) ;
endfunction


